"""
Custom exceptions for sportsdataverse module
"""


class SeasonNotFoundError(Exception):
    pass


def season_not_found_error(season, min_season):
    if int(season) >= int(min_season):
        return
    else:
        raise SeasonNotFoundError(f"Season {season} not found, season cannot be less than {min_season}")


class NoESPNDataError(Exception):
    pass


def no_espn_data(response):
    if response.status_code == 404:
        raise NoESPNDataError(f"NoESPNDataError: No response for {response.url}")
    elif response.json().get("code", None) == 404:
        raise NoESPNDataError(f"NoESPNDataError: No data found for {response.url}, response: {response.json()}")
    else:
        return response



import datetime

import pandas as pd
import polars as pl

from sportsdataverse.dl_utils import download


def espn_cfb_schedule(
    dates=None, week=None, season_type=None, groups=None, limit=500, return_as_pandas=False, **kwargs
) -> pl.DataFrame:
    """espn_cfb_schedule - look up the college football schedule for a given season


    Args:

        dates (int): Used to define different seasons. 2002 is the earliest available season.

        week (int): Week of the schedule.

        groups (int): Used to define different divisions. 80 is FBS, 81 is FCS.

        season_type (int): 2 for regular season, 3 for post-season, 4 for off-season.

        limit (int): number of records to return, default: 500.

        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.


    Returns:

        pl.DataFrame: Polars dataframe containing schedule dates for the requested season. Returns None if no games
    """

    params = {
        "week": week,
        "dates": dates,
        "seasonType": season_type,
        "groups": groups if groups is not None else "80",
        "limit": limit,
    }

    url = "http://site.api.espn.com/apis/site/v2/sports/football/college-football/scoreboard"

    resp = download(url=url, params=params, **kwargs)

    ev = pd.DataFrame()

    events_txt = resp.json()

    events = events_txt.get("events")

    if events is None:
        return None

    if len(events) == 0:
        return None

    for event in events:
        event = scoreboard_event_parsing(event)
        x = pl.from_pandas(pd.json_normalize(event.get("competitions")[0], sep="_"))
        x = x.with_columns(
            game_id=(pl.col("id").cast(pl.Int32)),
            season=(event.get("season").get("year")),
            season_type=(event.get("season").get("type")),
            week=(event.get("week", {}).get("number")),
            home_linescores=pl.when(pl.col("status_type_description").is_in(["Postponed", "Canceled"]))
            .then(None)
            .otherwise(pl.col("home_linescores")),
            away_linescores=pl.when(pl.col("status_type_description").is_in(["Postponed", "Canceled"]))
            .then(None)
            .otherwise(pl.col("away_linescores")),
        ).with_columns(
            season=pl.col("season").cast(pl.Int32),
            season_type=pl.col("season_type").cast(pl.Int32),
            week=pl.col("week").cast(pl.Int32),
        )

        x = x[[s.name for s in x if s.null_count() != x.height]]

        ev = pd.concat([ev, x.to_pandas()], axis=0, ignore_index=True)
    ev = pl.from_pandas(ev)
    ev = ev.janitor.clean_names()

    return ev.to_pandas() if return_as_pandas else ev


def scoreboard_event_parsing(event):
    event.get("competitions")[0].get("competitors")[0].get("team").pop("links", None)
    event.get("competitions")[0].get("competitors")[1].get("team").pop("links", None)
    if event.get("competitions")[0].get("competitors")[0].get("homeAway") == "home":
        event = __extract_home_away(event, 0, "home")
        event = __extract_home_away(event, 1, "away")
    else:
        event = __extract_home_away(event, 0, "away")
        event = __extract_home_away(event, 1, "home")
    del_keys = ["geoBroadcasts", "headlines", "series", "situation", "tickets", "odds", "leaders"]
    for k in del_keys:
        event.get("competitions")[0].pop(k, None)
    event.get("competitions")[0]["notes_type"] = (
        event.get("competitions")[0]["notes"][0].get("type") if len(event.get("competitions")[0]["notes"]) > 0 else ""
    )
    event.get("competitions")[0]["notes_headline"] = (
        event.get("competitions")[0]["notes"][0].get("headline").replace('"', "")
        if len(event.get("competitions")[0]["notes"]) > 0
        else ""
    )
    event.get("competitions")[0]["broadcast_market"] = (
        event.get("competitions")[0].get("broadcasts", [])[0].get("market", "")
        if len(event.get("competitions")[0].get("broadcasts")) > 0
        else ""
    )
    event.get("competitions")[0]["broadcast_name"] = (
        event.get("competitions")[0].get("broadcasts", [])[0].get("names", [])[0]
        if len(event.get("competitions")[0].get("broadcasts")) > 0
        else ""
    )
    event.get("competitions")[0].pop("broadcasts", None)
    event.get("competitions")[0].pop("notes", None)
    event.get("competitions")[0].pop("competitors", None)
    return event


def __extract_home_away(event, arg1, arg2):
    event["competitions"][0][arg2] = event.get("competitions")[0].get("competitors")[arg1].get("team")

    event["competitions"][0][arg2]["score"] = event.get("competitions")[0].get("competitors")[arg1].get("score")

    event["competitions"][0][arg2]["winner"] = event.get("competitions")[0].get("competitors")[arg1].get("winner")

    # add winner back to main competitors if does not exist

    event["competitions"][0]["competitors"][arg1]["winner"] = (
        event.get("competitions")[0].get("competitors")[arg1].get("winner", False)
    )

    event["competitions"][0][arg2]["currentRank"] = (
        event.get("competitions")[0].get("competitors")[arg1].get("curatedRank", {}).get("current", 99)
    )

    event["competitions"][0][arg2]["linescores"] = (
        event.get("competitions")[0]
        .get("competitors")[arg1]
        .get("linescores", [{"value": 0}, {"value": 0}, {"value": 0}, {"value": 0}])
    )

    # add linescores back to main competitors if does not exist

    event["competitions"][0]["competitors"][arg1]["linescores"] = (
        event.get("competitions")[0]
        .get("competitors")[arg1]
        .get("linescores", [{"value": 0}, {"value": 0}, {"value": 0}, {"value": 0}])
    )

    event["competitions"][0][arg2]["records"] = (
        event.get("competitions")[0]
        .get("competitors")[arg1]
        .get(
            "records",
            [
                {"abbreviation": "Game", "name": "overall", "summary": "0-0", "type": "total"},
                {"abbreviation": "null", "name": "Home", "summary": "0-0", "type": "home"},
                {"abbreviation": "null", "name": "Road", "summary": "0-0", "type": "road"},
                {"abbreviation": "null", "name": "vs. Conf.", "summary": "0-0", "type": "vsconf"},
            ],
        )
    )
    return event


def espn_cfb_calendar(season=None, groups=None, ondays=None, return_as_pandas=False, **kwargs) -> pl.DataFrame:
    """espn_cfb_calendar - look up the men's college football calendar for a given season


    Args:

        season (int): Used to define different seasons. 2002 is the earliest available season.

        groups (int): Used to define different divisions. 80 is FBS, 81 is FCS.

        ondays (boolean): Used to return dates for calendar ondays

        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.


    Returns:

        pl.DataFrame: Polars dataframe containing calendar dates for the requested season.


    Raises:

        ValueError: If `season` is less than 2002.
    """

    if ondays is not None:
        full_schedule = __ondays_cfb_calendar(season, **kwargs)

    else:
        url = "http://site.api.espn.com/apis/site/v2/sports/football/college-football/scoreboard"

        params = {"dates": season, "groups": groups if groups is not None else "80"}

        resp = download(url=url, params=params, **kwargs)

        txt = resp.json()

        txt = txt.get("leagues")[0].get("calendar")

        full_schedule = pl.DataFrame()

        for i in range(len(txt)):
            if txt[i].get("entries", None) is not None:
                reg = pd.json_normalize(
                    data=txt[i],
                    record_path="entries",
                    meta=["label", "value", "startDate", "endDate"],
                    meta_prefix="season_type_",
                    record_prefix="week_",
                    errors="ignore",
                    sep="_",
                )

                full_schedule = pl.concat([full_schedule, pl.from_pandas(reg)], how="vertical")

        full_schedule = full_schedule.with_columns(season=season)

        full_schedule = full_schedule.janitor.clean_names()

        full_schedule = full_schedule.rename({"week_value": "week", "season_type_value": "season_type"})

    return full_schedule.to_pandas() if return_as_pandas else full_schedule


def __ondays_cfb_calendar(season, **kwargs):
    url = f"https://sports.core.api.espn.com/v2/sports/football/leagues/college-football/seasons/{season}/types/2/calendar/ondays"

    resp = download(url=url, **kwargs)

    if resp is not None:
        txt = resp.json().get("eventDate").get("dates")

        result = pl.DataFrame(txt, schema=["dates"])

        result = result.with_columns(dateURL=pl.col("dates").str.slice(0, 10))

        result = result.with_columns(
            url="http://site.api.espn.com/apis/site/v2/sports/football/college-football/scoreboard?dates="
            + pl.col("dateURL")
        )

    return result


def most_recent_cfb_season():
    date = datetime.datetime.now()

    if date.month >= 8 and date.day >= 15:
        return date.year

    elif date.month >= 9:
        return date.year

    else:
        return date.year - 1



from sportsdataverse.nhl.nhl_api import *
from sportsdataverse.nhl.nhl_loaders import *
from sportsdataverse.nhl.nhl_pbp import *
from sportsdataverse.nhl.nhl_schedule import *
from sportsdataverse.nhl.nhl_teams import *



from sportsdataverse.wbb.wbb_game_rosters import *
from sportsdataverse.wbb.wbb_loaders import *
from sportsdataverse.wbb.wbb_pbp import *
from sportsdataverse.wbb.wbb_schedule import *
from sportsdataverse.wbb.wbb_teams import *



import datetime

import pandas as pd
import polars as pl

from sportsdataverse.dl_utils import download
from sportsdataverse.errors import SeasonNotFoundError


def espn_mbb_schedule(
    dates=None, groups=50, season_type=None, limit=500, return_as_pandas=False, **kwargs
) -> pl.DataFrame:
    """espn_mbb_schedule - look up the men's college basketball scheduler for a given season

    Args:
        dates (int): Used to define different seasons. 2002 is the earliest available season.
        groups (int): Used to define different divisions. 50 is Division I, 51 is Division II/Division III.
        season_type (int): 2 for regular season, 3 for post-season, 4 for off-season.
        limit (int): number of records to return, default: 500.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.
    Returns:
        pl.DataFrame: Polars dataframe containing schedule dates for the requested season. Returns None if no games
    """
    url = "http://site.api.espn.com/apis/site/v2/sports/basketball/mens-college-basketball/scoreboard"
    params = {
        "dates": dates,
        "seasonType": season_type,
        "groups": groups if groups is not None else "50",
        "limit": limit,
    }
    resp = download(url=url, params=params, **kwargs)

    ev = pd.DataFrame()
    events_txt = resp.json()
    events = events_txt.get("events")
    if events is None:
        return pd.DataFrame() if return_as_pandas else pl.DataFrame()
    if len(events) == 0:
        return pd.DataFrame() if return_as_pandas else pl.DataFrame()

    for event in events:
        event = scoreboard_event_parsing(event)
        x = pl.from_pandas(pd.json_normalize(event.get("competitions")[0], sep="_"))
        x = x.with_columns(
            game_id=(pl.col("id").cast(pl.Int32)),
            season=(event.get("season").get("year")),
            season_type=(event.get("season").get("type")),
            home_linescores=pl.when(pl.col("status_type_description") == "Postponed")
            .then(None)
            .otherwise(pl.col("home_linescores")),
            away_linescores=pl.when(pl.col("status_type_description") == "Postponed")
            .then(None)
            .otherwise(pl.col("away_linescores")),
        ).with_columns(
            season=pl.col("season").cast(pl.Int32),
            season_type=pl.col("season_type").cast(pl.Int32),
        )
        x = x[[s.name for s in x if s.null_count() != x.height]]
        ev = pd.concat([ev, x.to_pandas()], axis=0, ignore_index=True)
    ev = pl.from_pandas(ev)
    ev = ev.janitor.clean_names()

    return ev.to_pandas() if return_as_pandas else ev


def scoreboard_event_parsing(event):
    event.get("competitions")[0].get("competitors")[0].get("team").pop("links", None)
    event.get("competitions")[0].get("competitors")[1].get("team").pop("links", None)
    if event.get("competitions")[0].get("competitors")[0].get("homeAway") == "home":
        event = __extract_home_away(event, 0, "home")
        event = __extract_home_away(event, 1, "away")
    else:
        event = __extract_home_away(event, 0, "away")
        event = __extract_home_away(event, 1, "home")
    del_keys = ["geoBroadcasts", "headlines", "series", "situation", "tickets", "odds", "leaders"]
    for k in del_keys:
        event.get("competitions")[0].pop(k, None)
    event.get("competitions")[0]["notes_type"] = (
        event.get("competitions")[0]["notes"][0].get("type") if len(event.get("competitions")[0]["notes"]) > 0 else ""
    )
    event.get("competitions")[0]["notes_headline"] = (
        event.get("competitions")[0]["notes"][0].get("headline").replace('"', "")
        if len(event.get("competitions")[0]["notes"]) > 0
        else ""
    )
    event.get("competitions")[0]["broadcast_market"] = (
        event.get("competitions")[0].get("broadcasts", [])[0].get("market", "")
        if len(event.get("competitions")[0].get("broadcasts")) > 0
        else ""
    )
    event.get("competitions")[0]["broadcast_name"] = (
        event.get("competitions")[0].get("broadcasts", [])[0].get("names", [])[0]
        if len(event.get("competitions")[0].get("broadcasts")) > 0
        else ""
    )
    event.get("competitions")[0].pop("broadcasts", None)
    event.get("competitions")[0].pop("notes", None)
    event.get("competitions")[0].pop("competitors", None)
    return event


def __extract_home_away(event, arg1, arg2):
    event["competitions"][0][arg2] = event.get("competitions")[0].get("competitors")[arg1].get("team")
    event["competitions"][0][arg2]["score"] = event.get("competitions")[0].get("competitors")[arg1].get("score")
    event["competitions"][0][arg2]["winner"] = event.get("competitions")[0].get("competitors")[arg1].get("winner")
    # add winner back to main competitors if does not exist
    event["competitions"][0]["competitors"][arg1]["winner"] = (
        event.get("competitions")[0].get("competitors")[arg1].get("winner", False)
    )
    event["competitions"][0][arg2]["currentRank"] = (
        event.get("competitions")[0].get("competitors")[arg1].get("curatedRank", {}).get("current", 99)
    )
    event["competitions"][0][arg2]["linescores"] = (
        event.get("competitions")[0].get("competitors")[arg1].get("linescores", [{"value": 0}, {"value": 0}])
    )
    # add linescores back to main competitors if does not exist
    event["competitions"][0]["competitors"][arg1]["linescores"] = (
        event.get("competitions")[0].get("competitors")[arg1].get("linescores", [{"value": 0}, {"value": 0}])
    )
    event["competitions"][0][arg2]["records"] = (
        event.get("competitions")[0]
        .get("competitors")[arg1]
        .get(
            "records",
            [
                {"abbreviation": "Game", "name": "overall", "summary": "0-0", "type": "total"},
                {"abbreviation": "null", "name": "Home", "summary": "0-0", "type": "home"},
                {"abbreviation": "null", "name": "Road", "summary": "0-0", "type": "road"},
                {"abbreviation": "null", "name": "vs. Conf.", "summary": "0-0", "type": "vsconf"},
            ],
        )
    )
    return event


def espn_mbb_calendar(season=None, ondays=None, return_as_pandas=False, **kwargs) -> pl.DataFrame:
    """espn_mbb_calendar - look up the men's college basketball calendar for a given season

    Args:
        season (int): Used to define different seasons. 2002 is the earliest available season.
        ondays (boolean): Used to return dates for calendar ondays
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing calendar dates for the requested season.

    Raises:
        ValueError: If `season` is less than 2002.
    """
    if int(season) < 2002:
        raise SeasonNotFoundError("season cannot be less than 2002")
    if ondays is not None:
        full_schedule = __ondays_mbb_calendar(season, **kwargs)
    else:
        url = f"http://site.api.espn.com/apis/site/v2/sports/basketball/mens-college-basketball/scoreboard?dates={season}"
        resp = download(url=url, **kwargs)
        txt = resp.json().get("leagues")[0].get("calendar")
        datenum = list(map(lambda x: x[:10].replace("-", ""), txt))
        date = list(map(lambda x: x[:10], txt))
        year = list(map(lambda x: x[:4], txt))
        month = list(map(lambda x: x[5:7], txt))
        day = list(map(lambda x: x[8:10], txt))
        data = {
            "season": season,
            "datetime": txt,
            "date": date,
            "year": year,
            "month": month,
            "day": day,
            "dateURL": datenum,
        }
        full_schedule = pl.DataFrame(data)
        full_schedule = full_schedule.with_columns(
            url="http://site.api.espn.com/apis/site/v2/sports/basketball/mens-college-basketball/scoreboard?dates="
            + pl.col("dateURL")
        )
    return full_schedule.to_pandas() if return_as_pandas else full_schedule


def __ondays_mbb_calendar(season, **kwargs):
    url = f"https://sports.core.api.espn.com/v2/sports/basketball/leagues/mens-college-basketball/seasons/{season}/types/2/calendar/ondays"
    resp = download(url=url, **kwargs)
    txt = resp.json().get("eventDate").get("dates")
    result = pl.DataFrame(txt, schema=["dates"])
    result = result.with_columns(dateURL=pl.col("dates").str.slice(0, 10))
    result = result.with_columns(
        url="http://site.api.espn.com/apis/site/v2/sports/basketball/mens-college-basketball/scoreboard?dates="
        + pl.col("dateURL")
    )

    return result


def most_recent_mbb_season():
    if datetime.datetime.now().month >= 10:
        return datetime.datetime.now().year + 1
    else:
        return datetime.datetime.now().year



from sportsdataverse.mbb.mbb_game_rosters import *
from sportsdataverse.mbb.mbb_loaders import *
from sportsdataverse.mbb.mbb_pbp import *
from sportsdataverse.mbb.mbb_schedule import *
from sportsdataverse.mbb.mbb_teams import *



import json
from typing import Dict

import requests


def nfl_token_gen():
    url = "https://api.nfl.com/v1/reroute"

    # TODO: resolve if DNT or x-domain-id are necessary.  pulled them from chrome inspector
    payload = "grant_type=client_credentials"
    headers = {"DNT": "1", "x-domain-id": "100", "Content-Type": "application/x-www-form-urlencoded"}

    response = requests.request("POST", url, headers=headers, data=payload)

    return json.loads(response.content)["access_token"]


def nfl_headers_gen():
    token = nfl_token_gen()
    return {
        "Host": "api.nfl.com",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:105.0) Gecko/20100101 Firefox/105.0",
        "Accept": "*/*",
        "Accept-Language": "en-US,en;q=0.5",
        "Accept-Encoding": "gzip, deflate, br",
        "Referer": "https://www.nfl.com/",
        "authorization": f"Bearer {token}",
        "Connection": "keep-alive",
        "Sec-Fetch-Dest": "empty",
        "Sec-Fetch-Mode": "cors",
        "Sec-Fetch-Site": "same-site",
        "Sec-Fetch-User": "?1",
        "Pragma": "no-cache",
        "Cache-Control": "no-cache",
    }


def nfl_game_details(game_id=None, headers=None, raw=False) -> Dict:
    """nfl_game_details()
    Args:
        game_id (int): Game ID
    Returns:
        Dict: Dictionary of odds and props data with keys
    Example:
        `nfl_df = nfl_game_details(
        game_id = '7ae87c4c-d24c-11ec-b23d-d15a91047884'
        )`
    """
    if headers is None:
        headers = nfl_headers_gen()
    pbp_txt = {}
    summary_url = f"https://api.nfl.com/experience/v1/gamedetails/{game_id}"
    summary_resp = requests.get(summary_url, headers=headers)
    summary = summary_resp.json()

    incoming_keys_expected = [
        "attendance",
        "distance",
        "down",
        "gameClock",
        "goalToGo",
        "homePointsOvertime",
        "homePointsQ1",
        "homePointsQ2",
        "homePointsQ3",
        "homePointsQ4",
        "homePointsTotal",
        "homeTeam",
        "homeTimeoutsRemaining",
        "homeTimeoutsUsed",
        "id",
        "offset",
        "period",
        "phase",
        "playReview",
        "possessionTeam",
        "quarter",
        "redzone",
        "scoringSummaries",
        "stadium",
        "startTime",
        "totalOffset",
        "visitorPointsOvertime",
        "visitorPointsQ1",
        "visitorPointsQ2",
        "visitorPointsQ3",
        "visitorPointsQ4",
        "visitorPointsTotal",
        "visitorTeam",
        "visitorTimeoutsRemaining",
        "visitorTimeoutsUsed",
        "weather",
        "yardLine",
        "yardsToGo",
        "drives",
        "plays",
    ]
    dict_keys_expected = ["homeTeam", "possessionTeam", "visitorTeam", "weather"]
    array_keys_expected = ["scoringSummaries", "drives", "plays"]
    if raw == True:
        return summary

    for k in incoming_keys_expected:
        if k in summary.keys():
            pbp_txt[k] = summary.get(f"{k}")
        else:
            pbp_txt[k] = {} if k in dict_keys_expected else []
    return pbp_txt


def nfl_game_schedule(season=2021, season_type="REG", week=1, headers=None, raw=False) -> Dict:
    """nfl_game_schedule()
    Args:
        season (int): season
        season_type (str): season type - REG, POST
        week (int): week
    Returns:
        Dict: Dictionary of odds and props data with keys
    Example:
        `nfl_df = nfl_game_schedule(
            season = 2021, seasonType='REG', week=1
        )`
    """
    if headers is None:
        headers = nfl_headers_gen()
    params = {"season": season, "seasonType": season_type, "week": week}
    pbp_txt = {}
    summary_url = "https://api.nfl.com/experience/v1/games"
    summary_resp = requests.get(summary_url, headers=headers, params=params)
    summary = summary_resp.json()

    incoming_keys_expected = [
        "id",
        "homeTeam",
        "awayTeam",
        "category",
        "date",
        "time",
        "broadcastInfo",
        "neutralSite",
        "venue",
        "season",
        "seasonType",
        "status",
        "week",
        "weekType",
        "externalIds",
        "ticketUrl",
        "ticketVendors",
        "detail",
    ]
    dict_keys_expected = ["homeTeam", "possessionTeam", "visitorTeam", "weather"]
    array_keys_expected = ["scoringSummaries", "drives", "plays"]
    return summary



from functools import lru_cache

import pandas as pd
import polars as pl

from sportsdataverse.dl_utils import download, underscore


@lru_cache(maxsize=None)
def espn_cfb_teams(groups=None, return_as_pandas=False, **kwargs) -> pl.DataFrame:
    """espn_cfb_teams - look up the college football teams

    Args:
        groups (int): Used to define different divisions. 80 is FBS, 81 is FCS.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing schedule dates for the requested season.
        This function caches by default, so if you want to refresh the data, use the command
        sportsdataverse.cfb.espn_cfb_teams.clear_cache().

    Example:
        `cfb_df = sportsdataverse.cfb.espn_cfb_teams()`

    """
    url = "http://site.api.espn.com/apis/site/v2/sports/football/college-football/teams"
    params = {"groups": groups if groups is not None else "80", "limit": 1000}
    resp = download(url=url, params=params, **kwargs)
    if resp is not None:
        events_txt = resp.json()

        teams = events_txt.get("sports")[0].get("leagues")[0].get("teams")
        del_keys = ["record", "links"]
        for team in teams:
            for k in del_keys:
                team.get("team").pop(k, None)
        teams = pd.json_normalize(teams, sep="_")
    teams.columns = [underscore(c) for c in teams.columns.tolist()]
    return teams if return_as_pandas else pl.from_pandas(teams)



from typing import List

import polars as pl
from tqdm import tqdm

from sportsdataverse.config import (
    CFB_BASE_URL,
    CFB_BETTING_LINES_URL,
    CFB_ROSTER_URL,
    CFB_TEAM_INFO_URL,
    CFB_TEAM_LOGO_URL,
    CFB_TEAM_SCHEDULE_URL,
)
from sportsdataverse.errors import SeasonNotFoundError


def load_cfb_pbp(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load college football play by play data going back to 2003

    Example:
        `cfb_df = sportsdataverse.cfb.load_cfb_pbp(seasons=range(2003,2021))`

    Args:
        seasons (list): Used to define different seasons. 2003 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing the play-by-plays available for the requested seasons.

    Raises:
        ValueError: If `season` is less than 2003.
    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        if int(i) < 2003:
            raise SeasonNotFoundError("season cannot be less than 2003")
        i_data = pl.read_parquet(CFB_BASE_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_cfb_schedule(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load college football schedule data

    Example:
        `cfb_df = sportsdataverse.cfb.load_cfb_schedule(seasons=range(2002,2021))`

    Args:
        seasons (list): Used to define different seasons. 2002 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing the schedule for the requested seasons.

    Raises:
        ValueError: If `season` is less than 2002.
    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        if int(i) < 2002:
            raise SeasonNotFoundError("season cannot be less than 2002")
        i_data = pl.read_parquet(CFB_TEAM_SCHEDULE_URL.format(season=i), use_pyarrow=True, columns=None)
        # data = data.append(i_data)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_cfb_rosters(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load roster data

    Example:
        `cfb_df = sportsdataverse.cfb.load_cfb_rosters(seasons=range(2014,2021))`

    Args:
        seasons (list): Used to define different seasons. 2014 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing rosters available for the requested seasons.

    Raises:
        ValueError: If `season` is less than 2014.
    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        if int(i) < 2004:
            raise SeasonNotFoundError("season cannot be less than 2004")
        i_data = pl.read_parquet(CFB_ROSTER_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_cfb_team_info(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load college football team info

    Example:
        `cfb_df = sportsdataverse.cfb.load_cfb_team_info(seasons=range(2002,2021))`

    Args:
        seasons (list): Used to define different seasons. 2002 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing the team info available for the requested seasons.

    Raises:
        ValueError: If `season` is less than 2002.
    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        if int(i) < 2002:
            raise SeasonNotFoundError("season cannot be less than 2002")
        try:
            i_data = pl.read_parquet(CFB_TEAM_INFO_URL.format(season=i), use_pyarrow=True, columns=None)
        except Exception:
            print(f"We don't seem to have data for the {i} season.")
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_cfb_betting_lines(return_as_pandas=False) -> pl.DataFrame:
    """Load college football betting lines information

    Example:
        `cfb_df = sportsdataverse.cfb.load_cfb_betting_lines()`

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing betting lines available for the available seasons.
    """

    return (
        pl.read_parquet(CFB_BETTING_LINES_URL, use_pyarrow=True, columns=None).to_pandas(
            use_pyarrow_extension_array=True
        )
        if return_as_pandas
        else pl.read_parquet(CFB_BETTING_LINES_URL, use_pyarrow=True, columns=None)
    )


def get_cfb_teams(return_as_pandas=False) -> pl.DataFrame:
    """Load college football team ID information and logos

    Example:
        `cfb_df = sportsdataverse.cfb.get_cfb_teams()`

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing teams available.
    """

    return (
        pl.read_parquet(CFB_TEAM_LOGO_URL, use_pyarrow=True, columns=None).to_pandas(use_pyarrow_extension_array=True)
        if return_as_pandas
        else pl.read_parquet(CFB_TEAM_LOGO_URL, use_pyarrow=True, columns=None)
    )



import functools
import os
import time
from functools import wraps

import psutil


def timer(number=10):
    """Decorator that times the function it wraps over repeated executions

    Args:
        number: int, The number of repeated executions of the function being wrapped

    Returns:
        func
    """

    def actual_wrapper(func):
        @functools.wraps(func)
        def wrapper_timer(*args, **kwargs):
            tic = time.perf_counter()
            for i in range(number - 1):
                func(*args, **kwargs)
            else:
                value = func(*args, **kwargs)
            toc = time.perf_counter()
            elapsed_time = toc - tic
            print(f"Elapsed time of {func.__name__} for {number} runs:\n" f" {elapsed_time:0.6f} seconds")
            return value

        return wrapper_timer

    return actual_wrapper


# this decorator is used to record memory usage of the decorated function
def record_mem_usage(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        process = psutil.Process(os.getpid())
        mem_start = process.memory_info()[0]
        rt = func(*args, **kwargs)
        mem_end = process.memory_info()[0]
        diff_KB = (mem_end - mem_start) // 1000
        print(f"memory usage of {func.__name__}: {diff_KB} KB")
        return rt

    return wrapper


def record_time_usage(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        print(f"Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds")
        return result

    return timeit_wrapper



import json
import logging
import re
import time
from itertools import chain, starmap

import numpy as np
import polars as pl
import requests

from sportsdataverse.errors import no_espn_data

logger = logging.getLogger("sdv.dl_utils")
logger.addHandler(logging.NullHandler())


def download(url, params=None, headers=None, proxy=None, timeout=30, num_retries=15, session=None, logger=None):
    session, params, logger = init_request_settings(params, session, logger)
    try:
        response = session.get(url, params=params, proxies=proxy, headers=headers, timeout=timeout)
        response = no_espn_data(response)

    except Exception as e:
        if num_retries == 0:
            logger.error(f"Retry Limit Exceeded: {url} \nparams: {params}\n {e}")

        if hasattr(e, "code") and getattr(e, "code") == 404:
            logger.error(f"404: {url} \nparams: {params}")

        if num_retries > 0:
            logger.warning("%s: %i - %s for url (%s)", e, response.status_code, response.reason, response.url)
            time.sleep(2)
            return download(
                url,
                params=params,
                proxy=proxy,
                headers=headers,
                timeout=timeout,
                num_retries=num_retries - 1,
                session=session,
                logger=logger,
            )
        else:
            logger.error(f"Download Error: {url} \nparams: {params}\n {e}")

    return response


def init_request_settings(params, session, logger):
    if params is None:
        params = {}

    if session is None:
        session = requests.Session()

    if logger is None:
        logger = logging.getLogger("sdv.dl_utils")
        logger.addHandler(logging.NullHandler())
    return session, params, logger


def flatten_json_iterative(dictionary, sep=".", ind_start=0):
    """Flattening a nested json file"""

    def unpack_one(parent_key, parent_value):
        """Unpack one level (only one) of nesting in json file"""

        # Unpacking one level

        if isinstance(parent_value, dict):
            for key, value in parent_value.items():
                t1 = parent_key + sep + key

                yield t1, value

        elif isinstance(parent_value, list):
            i = ind_start

            for value in parent_value:
                t2 = parent_key + sep + str(i)

                i += 1

                yield t2, value
        else:
            yield parent_key, parent_value

    # Continue iterating the unpack_one function until the terminating condition is satisfied

    while True:
        # Continue unpacking the json file until all values are atomic elements (aka neither a dictionary nor a list)

        dictionary = dict(chain.from_iterable(starmap(unpack_one, dictionary.items())))

        # Terminating condition: none of the values in the json file are a dictionary or a list

        if not any(isinstance(value, dict) for value in dictionary.values()) and not any(
            isinstance(value, list) for value in dictionary.values()
        ):
            break

    return dictionary


def key_check(obj, key, replacement=np.array([])):
    return obj[key] if key in obj.keys() else replacement


@pl.api.register_dataframe_namespace("janitor")
class ColumnJanitor:
    def __init__(self, df: pl.DataFrame):
        self._df = df

    def clean_names(self) -> pl.DataFrame:
        return self._df.rename({c: underscore(c) for c in self._df.columns})

    def to_pascal_case(self) -> pl.DataFrame:
        return self._df.rename({c: camelize(c, True) for c in self._df.columns})

    def to_camel_case(self) -> pl.DataFrame:
        return self._df.rename({c: camelize(c, False) for c in self._df.columns})

    def to_kebab_case(self) -> pl.DataFrame:
        return self._df.rename({c: kebabize(c) for c in self._df.columns})


def underscore(word):
    """

    Make an underscored, lowercase form from the expression in the string.


    Example::


        >>> underscore("DeviceType")

        'device_type'


    As a rule of thumb you can think of :func:`underscore` as the inverse of

    :func:`camelize`, though there are cases where that does not hold::


        >>> camelize(underscore("IOError"))

        'IoError'


    """

    word = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", word)

    word = re.sub(r"([a-z\d])([A-Z])", r"\1_\2", word)

    word = word.replace("-", "_")

    return word.lower()


def kebabize(word):
    """

    Make a kebab-case, lowercase form from the expression in the string.


    Example::


        >>> kebabize("DeviceType")

        'device-type'


    As a rule of thumb you can think of :func:`kebabize` as the sister function of

    :func:`underscore`, replacing underscores with dashes

    """

    word = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", word)

    word = re.sub(r"([a-z\d])([A-Z])", r"\1_\2", word)

    word = word.replace("_", "-")

    return word.lower()


def camelize(string, uppercase_first_letter=True):
    """

    Convert strings to CamelCase.


    Examples::


        >>> camelize("device_type")

        'DeviceType'

        >>> camelize("device_type", False)

        'deviceType'


    :func:`camelize` can be thought of as a inverse of :func:`underscore`,

    although there are some cases where that does not hold::


        >>> camelize(underscore("IOError"))

        'IoError'


    :param uppercase_first_letter: if set to `True` :func:`camelize` converts

        strings to UpperCamelCase. If set to `False` :func:`camelize` produces

        lowerCamelCase. Defaults to `True`.

    """
    if uppercase_first_letter:
        return re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), string)
    else:
        return string[0].lower() + camelize(string)[1:]


class ESPNResponse:
    def __init__(self, response, status_code, url):
        self._response = response

        self._status_code = status_code

        self._url = url

    def get_response(self):
        return self._response

    def get_dict(self):
        return json.loads(self._response)

    def get_json(self):
        return json.dumps(self.get_dict())

    def valid_json(self):
        try:
            self.get_dict()

        except ValueError:
            return False

        return True

    def get_url(self):
        return self._url


class ESPNHTTP:
    espn_response = ESPNResponse

    base_url = None

    parameters = None

    headers = None

    def clean_contents(self, contents):
        return contents

    def send_api_request(
        self, endpoint, parameters, referer=None, headers=None, timeout=None, raise_exception_on_error=False
    ):
        if not self.base_url:
            raise Exception("Cannot use send_api_request from _HTTP class.")

        base_url = self.base_url.format(endpoint=endpoint)

        endpoint = endpoint.lower()

        self.parameters = parameters

        request_headers = self.headers if headers is None else headers
        if referer:
            request_headers["Referer"] = referer

        url = None

        status_code = None

        contents = None

        # Sort parameters by key... for some reason this matters for some requests...

        parameters = sorted(parameters.items(), key=lambda kv: kv[0])

        if not contents:
            response = requests.get(url=base_url, params=parameters, headers=request_headers, timeout=timeout)

            url = response.url

            status_code = response.status_code

            contents = response.text

        contents = self.clean_contents(contents)

        data = self.espn_response(response=contents, status_code=status_code, url=url)

        if raise_exception_on_error and not data.valid_json():
            raise Exception("InvalidResponse: Response is not in a valid JSON format.")

        return data



import tempfile
from typing import List

import polars as pl
from pyreadr import download_file, read_r
from tqdm import tqdm

from sportsdataverse.config import (
    NFL_BASE_URL,
    NFL_COMBINE_URL,
    NFL_CONTRACTS_URL,
    NFL_DEPTH_CHARTS_URL,
    NFL_DRAFT_PICKS_URL,
    NFL_INJURIES_URL,
    NFL_NGS_PASSING_URL,
    NFL_NGS_RECEIVING_URL,
    NFL_NGS_RUSHING_URL,
    NFL_OFFICIALS_URL,
    NFL_PBP_PARTICIPATION_URL,
    NFL_PFR_SEASON_DEF_URL,
    NFL_PFR_SEASON_PASS_URL,
    NFL_PFR_SEASON_REC_URL,
    NFL_PFR_SEASON_RUSH_URL,
    NFL_PFR_WEEK_DEF_URL,
    NFL_PFR_WEEK_PASS_URL,
    NFL_PFR_WEEK_REC_URL,
    NFL_PFR_WEEK_RUSH_URL,
    NFL_PLAYER_KICKING_STATS_URL,
    NFL_PLAYER_STATS_URL,
    NFL_PLAYER_URL,
    NFL_ROSTER_URL,
    NFL_SNAP_COUNTS_URL,
    NFL_TEAM_LOGO_URL,
    NFL_TEAM_SCHEDULE_URL,
    NFL_WEEKLY_ROSTER_URL,
)
from sportsdataverse.errors import season_not_found_error


def load_nfl_pbp(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL play by play data going back to 1999

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_pbp(seasons=range(1999,2021))`

    Args:
        seasons (list): Used to define different seasons. 1999 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing the play-by-plays available for the requested seasons.

    Raises:
        ValueError: If `season` is less than 1999.
    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 1999)
        i_data = pl.read_parquet(NFL_BASE_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_schedule(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL schedule data

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_schedule(seasons=range(1999,2021))`

    Args:
        seasons (list): Used to define different seasons. 1999 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing the schedule for the requested seasons.

    Raises:
        ValueError: If `season` is less than 1999.
    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    with tempfile.TemporaryDirectory() as tempdirname:
        for i in tqdm(seasons):
            season_not_found_error(int(i), 1999)
            schedule_url = NFL_TEAM_SCHEDULE_URL.format(season=i)
            # i_data = pd.read_parquet(NFL_TEAM_SCHEDULE_URL.format(season = i), engine='auto', columns=None)
            i_data = read_r(download_file(schedule_url, f"{tempdirname}/nfl_sched_{i}.rds"))[None]
            i_data = pl.DataFrame(i_data)
            data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_player_stats(kicking=False, return_as_pandas=False) -> pl.DataFrame:
    """Load NFL player stats data

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_player_stats()`

    Args:
        kicking (bool): If True, load kicking stats. If False, load all other stats.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing player stats.
    """
    data = pl.DataFrame()
    if kicking is False:
        data = pl.read_parquet(NFL_PLAYER_STATS_URL, use_pyarrow=True, columns=None)
    else:
        data = pl.read_parquet(NFL_PLAYER_KICKING_STATS_URL, use_pyarrow=True, columns=None)

    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_ngs_passing(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL NextGen Stats Passing data going back to 2016

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_ngs_passing()`

    Returns:
        pl.DataFrame: Polars dataframe containing the NextGen Stats Passing data available.

    """
    return (
        pl.read_parquet(NFL_NGS_PASSING_URL, use_pyarrow=True, columns=None).to_pandas(
            use_pyarrow_extension_array=True
        )
        if return_as_pandas
        else pl.read_parquet(NFL_NGS_PASSING_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_ngs_rushing(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL NextGen Stats Rushing data going back to 2016

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_ngs_rushing()`

    Returns:
        pl.DataFrame: Polars dataframe containing the NextGen Stats Rushing data available.

    """
    return (
        pl.read_parquet(NFL_NGS_RUSHING_URL, use_pyarrow=True, columns=None).to_pandas(
            use_pyarrow_extension_array=True
        )
        if return_as_pandas
        else pl.read_parquet(NFL_NGS_RUSHING_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_ngs_receiving(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL NextGen Stats Receiving data going back to 2016

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_ngs_receiving()`

    Returns:
        pl.DataFrame: Polars dataframe containing the NextGen Stats Receiving data available.

    """
    return (
        pl.read_parquet(NFL_NGS_RECEIVING_URL, use_pyarrow=True, columns=None).to_pandas(
            use_pyarrow_extension_array=True
        )
        if return_as_pandas
        else pl.read_parquet(NFL_NGS_RECEIVING_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_pfr_pass(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Pro-Football Reference Advanced Passing data going back to 2018

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_pfr_pass()`

    Returns:
        pl.DataFrame: Polars dataframe containing Pro-Football Reference
            advanced passing stats data available.

    """
    return (
        pl.read_parquet(NFL_PFR_SEASON_PASS_URL, use_pyarrow=True, columns=None).to_pandas(
            use_pyarrow_extension_array=True
        )
        if return_as_pandas
        else pl.read_parquet(NFL_PFR_SEASON_PASS_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_pfr_weekly_pass(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Pro-Football Reference Weekly Advanced Passing data going back to 2018

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_pfr_weekly_pass(seasons=range(2018,2021))`

    Args:
        seasons (list): Used to define different seasons. 2018 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing Pro-Football Reference
            advanced passing stats data available for the requested seasons.

    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 2018)
        i_data = pl.read_parquet(NFL_PFR_WEEK_PASS_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_pfr_rush(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Pro-Football Reference Advanced Rushing data going back to 2018

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_pfr_rush()`

    Returns:
        pl.DataFrame: Polars dataframe containing Pro-Football Reference
            advanced rushing stats data available.

    """
    return (
        pl.read_parquet(NFL_PFR_SEASON_RUSH_URL, use_pyarrow=True, columns=None).to_pandas(
            use_pyarrow_extension_array=True
        )
        if return_as_pandas
        else pl.read_parquet(NFL_PFR_SEASON_RUSH_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_pfr_weekly_rush(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Pro-Football Reference Weekly Advanced Rushing data going back to 2018

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_pfr_weekly_rush(seasons=range(2018,2021))`

    Args:
        seasons (list): Used to define different seasons. 2018 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing Pro-Football Reference
            advanced rushing stats data available for the requested seasons.

    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 2018)
        i_data = pl.read_parquet(NFL_PFR_WEEK_RUSH_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_pfr_rec(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Pro-Football Reference Advanced Receiving data going back to 2018

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_pfr_rec()`

    Returns:
        pl.DataFrame: Polars dataframe containing Pro-Football Reference
            advanced receiving stats data available.

    """
    return (
        pl.read_parquet(NFL_PFR_SEASON_REC_URL, use_pyarrow=True, columns=None).to_pandas(
            use_pyarrow_extension_array=True
        )
        if return_as_pandas
        else pl.read_parquet(NFL_PFR_SEASON_REC_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_pfr_weekly_rec(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Pro-Football Reference Weekly Advanced Receiving data going back to 2018

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_pfr_weekly_rec(seasons=range(2018,2021))`

    Args:
        seasons (list): Used to define different seasons. 2018 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing Pro-Football Reference
            advanced receiving stats data available for the requested seasons.

    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 2018)
        i_data = pl.read_parquet(NFL_PFR_WEEK_REC_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_pfr_def(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Pro-Football Reference Advanced Defensive data going back to 2018

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_pfr_def()`

    Returns:
        pl.DataFrame: Polars dataframe containing Pro-Football Reference
            advanced defensive stats data available.

    """
    return (
        pl.read_parquet(NFL_PFR_SEASON_DEF_URL, use_pyarrow=True, columns=None).to_pandas(
            use_pyarrow_extension_array=True
        )
        if return_as_pandas
        else pl.read_parquet(NFL_PFR_SEASON_DEF_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_pfr_weekly_def(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Pro-Football Reference Weekly Advanced Defensive data going back to 2018

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_pfr_weekly_def(seasons=range(2018,2021))`

    Args:
        seasons (list): Used to define different seasons. 2018 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing Pro-Football Reference
            advanced defensive stats data available for the requested seasons.

    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 2018)
        i_data = pl.read_parquet(NFL_PFR_WEEK_DEF_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_rosters(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL roster data for all seasons

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_rosters(seasons=range(1999,2021))`

    Args:
        seasons (list): Used to define different seasons. 1920 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing rosters available for the requested seasons.

    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 1920)
        i_data = pl.read_parquet(NFL_ROSTER_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_weekly_rosters(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL weekly roster data for selected seasons

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_weekly_rosters(seasons=range(2002,2021))`

    Args:
        seasons (list): Used to define different seasons. 2002 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing weekly rosters available for the requested seasons.

    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 2002)
        i_data = pl.read_parquet(NFL_WEEKLY_ROSTER_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_teams(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL team ID information and logos

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_teams()`

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing teams available.
    """
    return (
        pl.read_csv(NFL_TEAM_LOGO_URL).to_pandas(use_pyarrow_extension_array=True)
        if return_as_pandas
        else pl.read_csv(NFL_TEAM_LOGO_URL)
    )


def load_nfl_players(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Player ID information

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_players()`

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing players available.
    """
    return (
        pl.read_parquet(NFL_PLAYER_URL, use_pyarrow=True, columns=None).to_pandas(use_pyarrow_extension_array=True)
        if return_as_pandas
        else pl.read_parquet(NFL_OFFICIALS_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_snap_counts(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL snap counts data for selected seasons

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_snap_counts(seasons=range(2012,2021))`

    Args:
        seasons (list): Used to define different seasons. 2012 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing snap counts available for the requested seasons.

    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 2012)
        i_data = pl.read_parquet(NFL_SNAP_COUNTS_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_pbp_participation(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL play-by-play participation data for selected seasons

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_pbp_participation(seasons=range(2016,2021))`

    Args:
        seasons (list): Used to define different seasons. 2016 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing play-by-play participation data available for the requested seasons.

    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 2016)
        i_data = pl.read_parquet(NFL_PBP_PARTICIPATION_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_injuries(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL injuries data for selected seasons

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_injuries(seasons=range(2009,2021))`

    Args:
        seasons (list): Used to define different seasons. 2009 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing injuries data available for the requested seasons.

    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 2009)
        i_data = pl.read_parquet(NFL_INJURIES_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_depth_charts(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Depth Chart data for selected seasons

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_depth_charts(seasons=range(2001,2021))`

    Args:
        seasons (list): Used to define different seasons. 2001 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing depth chart data available for the requested seasons.

    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        season_not_found_error(int(i), 2001)
        i_data = pl.read_parquet(NFL_DEPTH_CHARTS_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_nfl_contracts(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Historical contracts information

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_contracts()`

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing historical contracts available.
    """
    return (
        pl.read_parquet(NFL_CONTRACTS_URL, use_pyarrow=True, columns=None).to_pandas(use_pyarrow_extension_array=True)
        if return_as_pandas
        else pl.read_parquet(NFL_CONTRACTS_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_combine(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Combine information

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_combine()`

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing NFL combine data available.
    """
    return (
        pl.read_parquet(NFL_COMBINE_URL, use_pyarrow=True, columns=None).to_pandas(use_pyarrow_extension_array=True)
        if return_as_pandas
        else pl.read_parquet(NFL_COMBINE_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_draft_picks(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Draft picks information

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_draft_picks()`

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing NFL Draft picks data available.
    """
    return (
        pl.read_parquet(NFL_DRAFT_PICKS_URL, use_pyarrow=True, columns=None).to_pandas(
            use_pyarrow_extension_array=True
        )
        if return_as_pandas
        else pl.read_parquet(NFL_DRAFT_PICKS_URL, use_pyarrow=True, columns=None)
    )


def load_nfl_officials(return_as_pandas=False) -> pl.DataFrame:
    """Load NFL Officials information

    Example:
        `nfl_df = sportsdataverse.nfl.load_nfl_officials()`

    Args:
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing officials available.
    """
    return (
        pl.read_parquet(NFL_OFFICIALS_URL, use_pyarrow=True, columns=None).to_pandas(use_pyarrow_extension_array=True)
        if return_as_pandas
        else pl.read_parquet(NFL_OFFICIALS_URL, use_pyarrow=True, columns=None)
    )


## Currently removed due to unsupported features of pyreadr's method.
## there is a list-column of nested tibbles within the data
## that is not supported by pyreadr
# def load_nfl_player_contracts_detail() -> pl.DataFrame:
#     """Load NFL Player contracts detail information

#     Example:
#         `nfl_df = sportsdataverse.nfl.load_nfl_player_contracts_detail()`

#     Args:

#     Returns:
#         pl.DataFrame: Polars dataframe containing player contracts detail data available.
#     """
#     data = pd.DataFrame()
#     with tempfile.TemporaryDirectory() as tempdirname:
#         df = read_r(
#             download_file(NFL_OTC_PLAYER_DETAILS_URL,
#                 "{}/otc_player_details.rds".format(tempdirname)))[None]
#         df = pd.DataFrame(df)
#         data = pd.concat([data, df], ignore_index=True)
#     #Give each row a unique index
#     data.reset_index(drop=True, inplace=True)
#     return df



from sportsdataverse.wnba.wnba_game_rosters import *
from sportsdataverse.wnba.wnba_loaders import *
from sportsdataverse.wnba.wnba_pbp import *
from sportsdataverse.wnba.wnba_schedule import *
from sportsdataverse.wnba.wnba_teams import *



from sportsdataverse.nfl.nfl_game_rosters import *
from sportsdataverse.nfl.nfl_games import *
from sportsdataverse.nfl.nfl_loaders import *
from sportsdataverse.nfl.nfl_pbp import *
from sportsdataverse.nfl.nfl_schedule import *
from sportsdataverse.nfl.nfl_teams import *



from typing import List

import polars as pl
from tqdm import tqdm

from sportsdataverse.config import (
    MBB_BASE_URL,
    MBB_PLAYER_BOX_URL,
    MBB_TEAM_BOX_URL,
    MBB_TEAM_SCHEDULE_URL,
)
from sportsdataverse.errors import SeasonNotFoundError


def load_mbb_pbp(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load men's college basketball play by play data going back to 2002

    Example:
        `mbb_df = sportsdataverse.mbb.load_mbb_pbp(seasons=range(2002,2022))`

    Args:
        seasons (list): Used to define different seasons. 2002 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing the
        play-by-plays available for the requested seasons.

    Raises:
        ValueError: If `season` is less than 2002.
    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        if int(i) < 2002:
            raise SeasonNotFoundError("season cannot be less than 2002")
        i_data = pl.read_parquet(MBB_BASE_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_mbb_team_boxscore(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load men's college basketball team boxscore data

    Example:
        `mbb_df = sportsdataverse.mbb.load_mbb_team_boxscore(seasons=range(2002,2022))`

    Args:
        seasons (list): Used to define different seasons. 2002 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing the
        team boxscores available for the requested seasons.

    Raises:
        ValueError: If `season` is less than 2002.
    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        if int(i) < 2002:
            raise SeasonNotFoundError("season cannot be less than 2002")
        i_data = pl.read_parquet(MBB_TEAM_BOX_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_mbb_player_boxscore(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load men's college basketball player boxscore data

    Example:
        `mbb_df = sportsdataverse.mbb.load_mbb_player_boxscore(seasons=range(2002,2022))`

    Args:
        seasons (list): Used to define different seasons. 2002 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing the
        player boxscores available for the requested seasons.

    Raises:
        ValueError: If `season` is less than 2002.
    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        if int(i) < 2002:
            raise SeasonNotFoundError("season cannot be less than 2002")
        i_data = pl.read_parquet(MBB_PLAYER_BOX_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data


def load_mbb_schedule(seasons: List[int], return_as_pandas=False) -> pl.DataFrame:
    """Load men's college basketball schedule data

    Example:
        `mbb_df = sportsdataverse.mbb.load_mbb_schedule(seasons=range(2002,2022))`

    Args:
        seasons (list): Used to define different seasons. 2002 is the earliest available season.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing the
        schedule for  the requested seasons.

    Raises:
        ValueError: If `season` is less than 2002.
    """
    data = pl.DataFrame()
    if type(seasons) is int:
        seasons = [seasons]
    for i in tqdm(seasons):
        if int(i) < 2002:
            raise SeasonNotFoundError("season cannot be less than 2002")
        i_data = pl.read_parquet(MBB_TEAM_SCHEDULE_URL.format(season=i), use_pyarrow=True, columns=None)
        data = pl.concat([data, i_data], how="vertical")
    return data.to_pandas(use_pyarrow_extension_array=True) if return_as_pandas else data



from sportsdataverse.cfb.cfb_game_rosters import *
from sportsdataverse.cfb.cfb_loaders import *
from sportsdataverse.cfb.cfb_pbp import *
from sportsdataverse.cfb.cfb_schedule import *
from sportsdataverse.cfb.cfb_teams import *



import pandas as pd
import polars as pl

from sportsdataverse.dl_utils import download, underscore


def espn_nhl_game_rosters(game_id: int, raw=False, return_as_pandas=False, **kwargs) -> pl.DataFrame:
    """espn_nhl_game_rosters() - Pull the game by id.

    Args:
        game_id (int): Unique game_id, can be obtained from espn_nhl_schedule().
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe of game roster data with columns:
        'athlete_id', 'athlete_uid', 'athlete_guid', 'athlete_type',
        'first_name', 'last_name', 'full_name', 'athlete_display_name',
        'short_name', 'weight', 'display_weight', 'height', 'display_height',
        'age', 'date_of_birth', 'slug', 'jersey', 'linked', 'active',
        'alternate_ids_sdr', 'birth_place_city', 'birth_place_state',
        'birth_place_country', 'headshot_href', 'headshot_alt',
        'experience_years', 'experience_display_value',
        'experience_abbreviation', 'status_id', 'status_name', 'status_type',
        'status_abbreviation', 'hand_type', 'hand_abbreviation',
        'hand_display_value', 'draft_display_text', 'draft_round', 'draft_year',
        'draft_selection', 'player_id', 'starter', 'valid', 'did_not_play',
        'display_name', 'ejected', 'athlete_href', 'position_href',
        'statistics_href', 'team_id', 'team_guid', 'team_uid', 'team_slug',
        'team_location', 'team_name', 'team_abbreviation',
        'team_display_name', 'team_short_display_name', 'team_color',
        'team_alternate_color', 'is_active', 'is_all_star',
        'logo_href', 'logo_dark_href', 'game_id'
    Example:
        `nhl_df = sportsdataverse.nhl.espn_nhl_game_rosters(game_id=401247153)`
    """
    # summary endpoint for pickcenter array
    summary_url = (
        "https://sports.core.api.espn.com/v2/sports/hockey/leagues/nhl/events/{x}/competitions/{x}/competitors".format(
            x=game_id
        )
    )
    summary_resp = download(summary_url, **kwargs)
    summary = summary_resp.json()
    items = helper_nhl_game_items(summary)
    team_rosters = helper_nhl_roster_items(items=items, summary_url=summary_url, **kwargs)
    team_rosters = team_rosters.join(items[["team_id", "order", "home_away", "winner"]], how="left", on="team_id")
    teams_df = helper_nhl_team_items(items=items, **kwargs)
    teams_rosters = team_rosters.join(teams_df, how="left", on="team_id")
    athletes = helper_nhl_athlete_items(teams_rosters=team_rosters, **kwargs)
    rosters = athletes.join(teams_rosters, how="left", left_on="athlete_id", right_on="player_id")
    rosters = rosters.with_columns(game_id=pl.lit(game_id).cast(pl.Int32))
    rosters.columns = [underscore(c) for c in rosters.columns]
    return rosters.to_pandas() if return_as_pandas else rosters


def helper_nhl_game_items(summary):
    items = pl.from_pandas(pd.json_normalize(summary, record_path="items", sep="_"))
    items.columns = [col.replace("$ref", "href") for col in items.columns]

    items.columns = [underscore(c) for c in items.columns]
    items = items.rename({"id": "team_id", "uid": "team_uid", "statistics_href": "team_statistics_href"})
    items = items.with_columns(team_id=pl.col("team_id").cast(pl.Int32))

    return items


def helper_nhl_team_items(items, **kwargs):
    pop_cols = [
        "$ref",
        "record",
        "athletes",
        "venue",
        "groups",
        "ranks",
        "statistics",
        "leaders",
        "links",
        "notes",
        "againstTheSpreadRecords",
        "franchise",
        "events",
        "college",
        "depthCharts",
        "transactions",
        "awards",
        "injuries",
        "coaches",
        "ranks",
        "attendance",
    ]
    teams_df = pl.DataFrame()
    for x in items["team_href"]:
        team = download(x, **kwargs).json()
        for k in pop_cols:
            team.pop(k, None)
        team_row = pl.from_pandas(pd.json_normalize(team, sep="_"))
        teams_df = pl.concat([teams_df, team_row], how="vertical")

    teams_df.columns = [
        "team_id",
        "team_guid",
        "team_uid",
        "team_slug",
        "team_location",
        "team_name",
        "team_nickname",
        "team_abbreviation",
        "team_display_name",
        "team_short_display_name",
        "team_color",
        "team_alternate_color",
        "is_active",
        "is_all_star",
        "logos",
        "team_alternate_ids_sdr",
    ]
    teams_df = teams_df.with_columns(logo_href=pl.lit(""), logo_dark_href=pl.lit(""))
    for row in range(len(teams_df["logos"])):
        team = teams_df["logos"][row]
        teams_df[row, "logo_href"] = team[0]["href"]
        teams_df[row, "logo_dark_href"] = team[1]["href"]

    teams_df = teams_df.drop(["logos"])
    teams_df = teams_df.with_columns(team_id=pl.col("team_id").cast(pl.Int32))
    return teams_df


def helper_nhl_roster_items(items, summary_url, **kwargs):
    team_ids = list(items["team_id"])
    game_rosters = pl.DataFrame()
    for tm in team_ids:
        team_roster_url = "{x}/{t}/roster".format(x=summary_url, t=tm)
        team_roster_resp = download(team_roster_url, **kwargs)
        team_roster = pl.from_pandas(pd.json_normalize(team_roster_resp.json().get("entries", []), sep="_"))
        team_roster.columns = [col.replace("$ref", "href") for col in team_roster.columns]
        team_roster.columns = [underscore(c) for c in team_roster.columns]
        team_roster = team_roster.with_columns(team_id=pl.lit(tm).cast(pl.Int32))
        game_rosters = pl.concat([game_rosters, team_roster], how="diagonal")
    game_rosters = game_rosters.with_columns(
        player_id=pl.col("player_id").cast(pl.Int64), team_id=pl.col("team_id").cast(pl.Int32)
    )
    return game_rosters


def helper_nhl_athlete_items(teams_rosters, **kwargs):
    athlete_hrefs = list(teams_rosters["athlete_href"])
    game_athletes = pl.DataFrame()
    pop_cols = [
        "links",
        "injuries",
        "teams",
        "team",
        "college",
        "proAthlete",
        "statistics",
        "notes",
        "eventLog",
        "$ref",
        "position",
    ]
    for athlete_href in athlete_hrefs:
        athlete_res = download(athlete_href, **kwargs)
        athlete_resp = athlete_res.json()
        for k in pop_cols:
            athlete_resp.pop(k, None)
        athlete = pl.from_pandas(pd.json_normalize(athlete_resp, sep="_"))
        athlete.columns = [col.replace("$ref", "href") for col in athlete.columns]
        athlete.columns = [underscore(c) for c in athlete.columns]

        game_athletes = pl.concat([game_athletes, athlete], how="diagonal")

    game_athletes = game_athletes.rename(
        {
            "id": "athlete_id",
            "uid": "athlete_uid",
            "guid": "athlete_guid",
            "type": "athlete_type",
            "display_name": "athlete_display_name",
        }
    )
    game_athletes = game_athletes.with_columns(athlete_id=pl.col("athlete_id").cast(pl.Int64))
    return game_athletes



from sportsdataverse.nba.nba_game_rosters import *
from sportsdataverse.nba.nba_loaders import *
from sportsdataverse.nba.nba_pbp import *
from sportsdataverse.nba.nba_schedule import *
from sportsdataverse.nba.nba_teams import *



import logging

logger = logging.getLogger("sdv.dl_utils")
logger.addHandler(logging.NullHandler())


SGITHUB = "https://raw.githubusercontent.com/sportsdataverse/"
SDVRELEASES = "https://github.com/sportsdataverse/sportsdataverse-data/releases/download/"

CFB_BASE_URL = SGITHUB + "cfbfastR-data/main/pbp/parquet/play_by_play_{season}.parquet"
CFB_ROSTER_URL = SGITHUB + "cfbfastR-data/main/rosters/parquet/cfb_rosters_{season}.parquet"
CFB_TEAM_LOGO_URL = f"{SGITHUB}cfbfastR-data/main/teams/teams_colors_logos.parquet"
CFB_TEAM_SCHEDULE_URL = SGITHUB + "cfbfastR-data/main/schedules/parquet/cfb_schedules_{season}.parquet"
CFB_TEAM_INFO_URL = SGITHUB + "cfbfastR-data/main/team_info/parquet/cfb_team_info_{season}.parquet"
CFB_BETTING_LINES_URL = f"{SGITHUB}cfbfastR-data/main/betting/parquet/cfb_line_odds.parquet"

NHL_BASE_URL = SGITHUB + "fastRhockey-data/main/nhl/pbp/parquet/play_by_play_{season}.parquet"
NHL_PLAYER_BOX_URL = SGITHUB + "fastRhockey-data/main/nhl/player_box/parquet/player_box_{season}.parquet"
NHL_TEAM_BOX_URL = SGITHUB + "fastRhockey-data/main/nhl/team_box/parquet/team_box_{season}.parquet"
NHL_TEAM_SCHEDULE_URL = SGITHUB + "fastRhockey-data/main/nhl/schedules/parquet/nhl_schedule_{season}.parquet"
NHL_TEAM_LOGO_URL = f"{SGITHUB}fastRhockey-data/main/nhl/nhl_teams_colors_logos.csv"

PHF_BASE_URL = SGITHUB + "fastRhockey-data/main/phf/pbp/parquet/play_by_play_{season}.parquet"
PHF_PLAYER_BOX_URL = SGITHUB + "fastRhockey-data/main/phf/player_box/parquet/player_box_{season}.parquet"
PHF_TEAM_BOX_URL = SGITHUB + "fastRhockey-data/main/phf/team_box/parquet/team_box_{season}.parquet"
PHF_TEAM_SCHEDULE_URL = SGITHUB + "fastRhockey-data/main/phf/schedules/parquet/phf_schedule_{season}.parquet"

MBB_BASE_URL = SDVRELEASES + "espn_mens_college_basketball_pbp/play_by_play_{season}.parquet"
MBB_TEAM_BOX_URL = SDVRELEASES + "espn_mens_college_basketball_team_boxscores/team_box_{season}.parquet"
MBB_PLAYER_BOX_URL = SDVRELEASES + "espn_mens_college_basketball_player_boxscores/player_box_{season}.parquet"
MBB_TEAM_LOGO_URL = f"{SDVRELEASES}hoopR-data/master/mbb/teams_colors_logos.csv"
MBB_TEAM_SCHEDULE_URL = SDVRELEASES + "espn_mens_college_basketball_schedules/mbb_schedule_{season}.parquet"

NBA_BASE_URL = SDVRELEASES + "espn_nba_pbp/play_by_play_{season}.parquet"
NBA_TEAM_BOX_URL = SDVRELEASES + "espn_nba_team_boxscores/team_box_{season}.parquet"
NBA_PLAYER_BOX_URL = SDVRELEASES + "espn_nba_player_boxscores/player_box_{season}.parquet"
NBA_TEAM_SCHEDULE_URL = SDVRELEASES + "espn_nba_schedules/nba_schedule_{season}.parquet"

WBB_BASE_URL = SDVRELEASES + "espn_womens_college_basketball_pbp/play_by_play_{season}.parquet"
WBB_TEAM_BOX_URL = SDVRELEASES + "espn_womens_college_basketball_team_boxscores/team_box_{season}.parquet"
WBB_PLAYER_BOX_URL = SDVRELEASES + "espn_womens_college_basketball_player_boxscores/player_box_{season}.parquet"
WBB_TEAM_LOGO_URL = f"{SDVRELEASES}wehoop-data/master/wbb/teams_colors_logos.csv"
WBB_TEAM_SCHEDULE_URL = SDVRELEASES + "espn_womens_college_basketball_schedules/wbb_schedule_{season}.parquet"

WNBA_BASE_URL = SDVRELEASES + "espn_wnba_pbp/play_by_play_{season}.parquet"
WNBA_TEAM_BOX_URL = SDVRELEASES + "espn_wnba_team_boxscores/team_box_{season}.parquet"
WNBA_PLAYER_BOX_URL = SDVRELEASES + "espn_wnba_player_boxscores/player_box_{season}.parquet"
WNBA_TEAM_SCHEDULE_URL = SDVRELEASES + "espn_wnba_schedules/wnba_schedule_{season}.parquet"


NFLVERSEGITHUB = "https://github.com/nflverse/nflverse-data/releases/download/"
NFLVERSEGITHUBPBP = "https://raw.githubusercontent.com/nflverse/"
NFL_BASE_URL = NFLVERSEGITHUB + "pbp/play_by_play_{season}.parquet"  # done
NFL_PLAYER_URL = f"{NFLVERSEGITHUB}players/players.parquet"
NFL_PLAYER_STATS_URL = f"{NFLVERSEGITHUB}player_stats/player_stats.parquet"
NFL_PLAYER_KICKING_STATS_URL = f"{NFLVERSEGITHUB}player_stats/player_stats_kicking.parquet"
NFL_PFR_SEASON_DEF_URL = f"{NFLVERSEGITHUB}pfr_advstats/advstats_season_def.parquet"
NFL_PFR_WEEK_DEF_URL = NFLVERSEGITHUB + "pfr_advstats/advstats_week_def_{season}.parquet"
NFL_PFR_SEASON_PASS_URL = f"{NFLVERSEGITHUB}pfr_advstats/advstats_season_pass.parquet"
NFL_PFR_WEEK_PASS_URL = NFLVERSEGITHUB + "pfr_advstats/advstats_week_pass_{season}.parquet"
NFL_PFR_SEASON_REC_URL = f"{NFLVERSEGITHUB}pfr_advstats/advstats_season_rec.parquet"
NFL_PFR_WEEK_REC_URL = NFLVERSEGITHUB + "pfr_advstats/advstats_week_rec_{season}.parquet"
NFL_PFR_SEASON_RUSH_URL = f"{NFLVERSEGITHUB}pfr_advstats/advstats_season_rush.parquet"
NFL_PFR_WEEK_RUSH_URL = NFLVERSEGITHUB + "pfr_advstats/advstats_week_rush_{season}.parquet"
NFL_NGS_RUSHING_URL = f"{NFLVERSEGITHUB}nextgen_stats/ngs_rushing.parquet"
NFL_NGS_PASSING_URL = f"{NFLVERSEGITHUB}nextgen_stats/ngs_passing.parquet"
NFL_NGS_RECEIVING_URL = f"{NFLVERSEGITHUB}nextgen_stats/ngs_receiving.parquet"
NFL_ROSTER_URL = NFLVERSEGITHUB + "rosters/roster_{season}.parquet"  # done
NFL_WEEKLY_ROSTER_URL = NFLVERSEGITHUB + "weekly_rosters/roster_weekly_{season}.parquet"  # done
NFL_SNAP_COUNTS_URL = NFLVERSEGITHUB + "snap_counts/snap_counts_{season}.parquet"
NFL_PBP_PARTICIPATION_URL = NFLVERSEGITHUB + "pbp_participation/pbp_participation_{season}.parquet"
NFL_CONTRACTS_URL = f"{NFLVERSEGITHUB}contracts/historical_contracts.parquet"
NFL_OTC_PLAYER_DETAILS_URL = f"{NFLVERSEGITHUB}contracts/otc_player_details.rds"
NFL_DRAFT_PICKS_URL = f"{NFLVERSEGITHUB}draft_picks/draft_picks.parquet"
NFL_COMBINE_URL = f"{NFLVERSEGITHUB}combine/combine.parquet"
NFL_INJURIES_URL = NFLVERSEGITHUB + "injuries/injuries_{season}.parquet"
NFL_DEPTH_CHARTS_URL = NFLVERSEGITHUB + "depth_charts/depth_charts_{season}.parquet"
NFL_OFFICIALS_URL = f"{NFLVERSEGITHUB}officials/officials.parquet"
NFL_TEAM_LOGO_URL = f"{NFLVERSEGITHUBPBP}nflverse-pbp/master/teams_colors_logos.csv"
NFL_TEAM_SCHEDULE_URL = NFLVERSEGITHUBPBP + "nflverse-pbp/master/schedules/sched_{season}.rds"



import json
import os
import re
from typing import Dict

import numpy as np
import pandas as pd
import polars as pl

from sportsdataverse.dl_utils import download, flatten_json_iterative, key_check


def espn_nhl_pbp(game_id: int, raw=False, **kwargs) -> Dict:
    """espn_nhl_pbp() - Pull the game by id. Data from API endpoints - `nhl/playbyplay`, `nhl/summary`

    Args:
        game_id (int): Unique game_id, can be obtained from nhl_schedule().

    Returns:
        Dict: Dictionary of game data with keys - "gameId", "plays", "boxscore", "header", "broadcasts",
         "videos", "playByPlaySource", "standings", "leaders", "seasonseries", "pickcenter", "againstTheSpread",
         "odds", "onIce", "gameInfo", "season"

    Example:
        `nhl_df = sportsdataverse.nhl.espn_nhl_pbp(game_id=401247153)`
    """
    pbp_txt = {}
    summary_url = f"http://site.api.espn.com/apis/site/v2/sports/hockey/nhl/summary?event={game_id}"
    summary_resp = download(summary_url, **kwargs)
    summary = summary_resp.json()
    for k in [
        "plays",
        "seasonseries",
        "videos",
        "broadcasts",
        "pickcenter",
        "onIce",
        "againstTheSpread",
        "odds",
        "winprobability",
        "teamInfo",
        "espnWP",
        "leaders",
    ]:
        pbp_txt[k] = key_check(obj=summary, key=k, replacement=np.array([]))
    for k in ["boxscore", "format", "gameInfo", "article", "header", "season", "standings"]:
        pbp_txt[k] = key_check(obj=summary, key=k, replacement={})
    for k in ["news", "shop"]:
        if k in pbp_txt.keys():
            del pbp_txt[k]
    incoming_keys_expected = [
        "boxscore",
        "format",
        "gameInfo",
        "leaders",
        "seasonseries",
        "broadcasts",
        "pickcenter",
        "againstTheSpread",
        "odds",
        "winprobability",
        "header",
        "onIce",
        "article",
        "videos",
        "plays",
        "standings",
        "teamInfo",
        "espnWP",
        "season",
        "timeouts",
    ]
    if raw == True:
        # reorder keys in raw format, appending empty keys which are defined later to the end
        pbp_json = {}
        for k in incoming_keys_expected:
            if k in pbp_txt.keys():
                pbp_json[k] = pbp_txt[k]
            else:
                pbp_json[k] = {}
        return pbp_json
    return helper_nhl_pbp(game_id, pbp_txt)


def nhl_pbp_disk(game_id, path_to_json):
    with open(os.path.join(path_to_json, f"{game_id}.json")) as json_file:
        pbp_txt = json.load(json_file)
    return pbp_txt


def helper_nhl_pbp(game_id, pbp_txt):
    init = helper_nhl_pickcenter(pbp_txt)
    pbp_txt, init = helper_nhl_game_data(pbp_txt, init)

    if (pbp_txt["playByPlaySource"] != "none") & (len(pbp_txt["plays"]) > 1):
        pbp_txt = helper_nhl_pbp_features(game_id, pbp_txt, init)
    else:
        pbp_txt["plays"] = pl.DataFrame()
    return {
        "gameId": game_id,
        "plays": pbp_txt["plays"].to_dicts(),
        "boxscore": pbp_txt["boxscore"],
        "header": pbp_txt["header"],
        "format": pbp_txt["format"],
        "broadcasts": np.array(pbp_txt["broadcasts"]).tolist(),
        "videos": np.array(pbp_txt["videos"]).tolist(),
        "playByPlaySource": pbp_txt["playByPlaySource"],
        "standings": pbp_txt["standings"],
        "leaders": np.array(pbp_txt["leaders"]).tolist(),
        "seasonseries": np.array(pbp_txt["seasonseries"]).tolist(),
        "pickcenter": np.array(pbp_txt["pickcenter"]).tolist(),
        "againstTheSpread": np.array(pbp_txt["againstTheSpread"]).tolist(),
        "odds": np.array(pbp_txt["odds"]).tolist(),
        "onIce": np.array(pbp_txt["onIce"]).tolist(),
        "gameInfo": pbp_txt["gameInfo"],
        "teamInfo": np.array(pbp_txt["teamInfo"]).tolist(),
        "season": np.array(pbp_txt["season"]).tolist(),
    }


def helper_nhl_game_data(pbp_txt, init):
    pbp_txt["timeouts"] = {}
    pbp_txt["teamInfo"] = pbp_txt["header"]["competitions"][0]
    pbp_txt["season"] = pbp_txt["header"]["season"]
    pbp_txt["playByPlaySource"] = pbp_txt["header"]["competitions"][0]["playByPlaySource"]
    pbp_txt["boxscoreSource"] = pbp_txt["header"]["competitions"][0]["boxscoreSource"]
    pbp_txt["gameSpreadAvailable"] = init["gameSpreadAvailable"]
    pbp_txt["gameSpread"] = init["gameSpread"]
    pbp_txt["homeFavorite"] = init["homeFavorite"]
    pbp_txt["homeTeamSpread"] = np.where(
        init["homeFavorite"] == True, abs(init["gameSpread"]), -1 * abs(init["gameSpread"])
    )
    pbp_txt["overUnder"] = init["overUnder"]
    # Home and Away identification variables
    if pbp_txt["header"]["competitions"][0]["competitors"][0]["homeAway"] == "home":
        pbp_txt["header"]["competitions"][0]["home"] = pbp_txt["header"]["competitions"][0]["competitors"][0]["team"]
        homeTeamId = int(pbp_txt["header"]["competitions"][0]["competitors"][0]["team"]["id"])
        homeTeamMascot = str(pbp_txt["header"]["competitions"][0]["competitors"][0]["team"]["name"])
        homeTeamName = str(pbp_txt["header"]["competitions"][0]["competitors"][0]["team"]["location"])
        homeTeamAbbrev = str(pbp_txt["header"]["competitions"][0]["competitors"][0]["team"]["abbreviation"])
        homeTeamNameAlt = re.sub("Stat(.+)", "St", homeTeamName)
        pbp_txt["header"]["competitions"][0]["away"] = pbp_txt["header"]["competitions"][0]["competitors"][1]["team"]
        awayTeamId = int(pbp_txt["header"]["competitions"][0]["competitors"][1]["team"]["id"])
        awayTeamMascot = str(pbp_txt["header"]["competitions"][0]["competitors"][1]["team"]["name"])
        awayTeamName = str(pbp_txt["header"]["competitions"][0]["competitors"][1]["team"]["location"])
        awayTeamAbbrev = str(pbp_txt["header"]["competitions"][0]["competitors"][1]["team"]["abbreviation"])
        awayTeamNameAlt = re.sub("Stat(.+)", "St", awayTeamName)
    else:
        pbp_txt["header"]["competitions"][0]["away"] = pbp_txt["header"]["competitions"][0]["competitors"][0]["team"]
        awayTeamId = int(pbp_txt["header"]["competitions"][0]["competitors"][0]["team"]["id"])
        awayTeamMascot = str(pbp_txt["header"]["competitions"][0]["competitors"][0]["team"]["name"])
        awayTeamName = str(pbp_txt["header"]["competitions"][0]["competitors"][0]["team"]["location"])
        awayTeamAbbrev = str(pbp_txt["header"]["competitions"][0]["competitors"][0]["team"]["abbreviation"])
        awayTeamNameAlt = re.sub("Stat(.+)", "St", awayTeamName)
        pbp_txt["header"]["competitions"][0]["home"] = pbp_txt["header"]["competitions"][0]["competitors"][1]["team"]
        homeTeamId = int(pbp_txt["header"]["competitions"][0]["competitors"][1]["team"]["id"])
        homeTeamMascot = str(pbp_txt["header"]["competitions"][0]["competitors"][1]["team"]["name"])
        homeTeamName = str(pbp_txt["header"]["competitions"][0]["competitors"][1]["team"]["location"])
        homeTeamAbbrev = str(pbp_txt["header"]["competitions"][0]["competitors"][1]["team"]["abbreviation"])
        homeTeamNameAlt = re.sub("Stat(.+)", "St", homeTeamName)
    init["homeTeamId"] = homeTeamId
    init["homeTeamMascot"] = homeTeamMascot
    init["homeTeamName"] = homeTeamName
    init["homeTeamAbbrev"] = homeTeamAbbrev
    init["homeTeamNameAlt"] = homeTeamNameAlt
    init["awayTeamId"] = awayTeamId
    init["awayTeamMascot"] = awayTeamMascot
    init["awayTeamName"] = awayTeamName
    init["awayTeamAbbrev"] = awayTeamAbbrev
    init["awayTeamNameAlt"] = awayTeamNameAlt
    return pbp_txt, init


def helper_nhl_pickcenter(pbp_txt):
    # Spread definition
    if len(pbp_txt.get("pickcenter", [])) > 1:
        pickcenter = pd.json_normalize(data=pbp_txt, record_path="pickcenter")
        pickcenter = pickcenter.sort_values(by=["provider.id"])
        homeFavorite = (
            pickcenter[pickcenter["homeTeamOdds.favorite"].notnull()][["homeTeamOdds.favorite"]].values[0]
            if "homeTeamOdds.favorite" in pickcenter.columns
            else True
        )
        gameSpread = (
            pickcenter[pickcenter["spread"].notnull()][["spread"]].values[0] if "spread" in pickcenter.columns else 1.5
        )
        overUnder = (
            pickcenter[pickcenter["overUnder"].notnull()][["overUnder"]].values[0]
            if "overUnder" in pickcenter.columns
            else 5.5
        )
        gameSpreadAvailable = True
        # self.logger.info(f"Spread: {gameSpread}, home Favorite: {homeFavorite}, ou: {overUnder}")
    else:
        gameSpread = 1.5
        overUnder = 5.5
        homeFavorite = True
        gameSpreadAvailable = False

    return {
        "gameSpread": gameSpread,
        "overUnder": overUnder,
        "homeFavorite": homeFavorite,
        "gameSpreadAvailable": gameSpreadAvailable,
    }


def helper_nhl_pbp_features(game_id, pbp_txt, init):
    pbp_txt["plays_mod"] = []
    for play in pbp_txt["plays"]:
        p = flatten_json_iterative(play)
        pbp_txt["plays_mod"].append(p)
    pbp_txt["plays"] = pl.from_pandas(pd.json_normalize(pbp_txt, "plays_mod"))
    pbp_txt["plays"] = (
        pbp_txt["plays"]
        .with_columns(
            game_id=pl.lit(game_id).cast(pl.Int32),
            id=(pl.col("id").cast(pl.Int64)),
            sequenceNumber=pl.col("sequenceNumber").cast(pl.Int32),
            season=pl.lit(pbp_txt["header"]["season"]["year"]).cast(pl.Int32),
            seasonType=pl.lit(pbp_txt["header"]["season"]["type"]),
            homeTeamId=pl.lit(init["homeTeamId"]).cast(pl.Int32),
            homeTeamName=pl.lit(init["homeTeamName"]),
            homeTeamMascot=pl.lit(init["homeTeamMascot"]),
            homeTeamAbbrev=pl.lit(init["homeTeamAbbrev"]),
            homeTeamNameAlt=pl.lit(init["homeTeamNameAlt"]),
            awayTeamId=pl.lit(init["awayTeamId"]).cast(pl.Int32),
            awayTeamName=pl.lit(init["awayTeamName"]),
            awayTeamMascot=pl.lit(init["awayTeamMascot"]),
            awayTeamAbbrev=pl.lit(init["awayTeamAbbrev"]),
            awayTeamNameAlt=pl.lit(init["awayTeamNameAlt"]),
            gameSpread=pl.lit(init["gameSpread"]).abs(),
            homeFavorite=pl.lit(init["homeFavorite"]),
            gameSpreadAvailable=pl.lit(init["gameSpreadAvailable"]),
        )
        .with_columns(
            homeTeamSpread=pl.when(pl.col("homeFavorite") == True)
            .then(pl.col("gameSpread"))
            .otherwise(-1 * pl.col("gameSpread")),
        )
        .with_columns(
            pl.col("period.number").cast(pl.Int32).alias("period.number"),
            pl.col("period.number").cast(pl.Int32).alias("qtr"),
            pl.when(pl.col("clock.displayValue").str.contains(r":") == False)
            .then("0:" + pl.col("clock.displayValue"))
            .otherwise(pl.col("clock.displayValue"))
            .alias("clock.displayValue"),
        )
        .with_columns(
            pl.col("clock.displayValue").alias("time"),
            pl.col("clock.displayValue")
            .str.split(":")
            .list.to_struct(n_field_strategy="first_non_null")
            .alias("clock.mm"),
        )
        .with_columns(pl.col("clock.mm").struct.rename_fields(["clock.minutes", "clock.seconds"]))
        .unnest("clock.mm")
        .with_columns(
            pl.col("clock.minutes").cast(pl.Float32),
            pl.col("clock.seconds").cast(pl.Float32),
        )
        .with_columns(
            lag_period=pl.col("period.number").shift(1),
            lead_period=pl.col("period.number").shift(-1),
        )
        .with_columns(
            (60 * pl.col("clock.minutes") + pl.col("clock.seconds")).alias("start.period_seconds_remaining"),
            pl.when(pl.col("period.number") == 1)
            .then(2400 + 60 * pl.col("clock.minutes") + pl.col("clock.seconds"))
            .when(pl.col("period.number") == 2)
            .then(1200 + 60 * pl.col("clock.minutes") + pl.col("clock.seconds"))
            .when(pl.col("period.number") == 3)
            .then(60 * pl.col("clock.minutes") + pl.col("clock.seconds"))
            .otherwise(60 * pl.col("clock.minutes") + pl.col("clock.seconds"))
            .alias("start.game_seconds_remaining"),
        )
        .with_columns(
            pl.col("start.period_seconds_remaining").shift(-1).alias("end.period_seconds_remaining"),
            pl.col("start.game_seconds_remaining").shift(-1).alias("end.game_seconds_remaining"),
        )
    )

    pbp_txt["plays"] = pbp_txt["plays"].with_row_count("game_play_number", 1)

    pbp_txt["plays"] = pbp_txt["plays"].with_columns(
        pl.when(pl.col("game_play_number") == 1)
        .then(1200)
        .when((pl.col("lag_period") == 1).and_(pl.col("period.number") == 2))
        .then(1200)
        .when((pl.col("lag_period") == 2).and_(pl.col("period.number") == 3))
        .then(1200)
        .when((pl.col("lag_period") == pl.col("period.number") - 1).and_(pl.col("period.number") >= 4))
        .then(1200)
        .otherwise(pl.col("end.period_seconds_remaining"))
        .alias("end.period_seconds_remaining"),
        pl.when(pl.col("game_play_number") == 1)
        .then(3600)
        .when((pl.col("lag_period") == 1).and_(pl.col("period.number") == 2))
        .then(2400)
        .when((pl.col("lag_period") == 2).and_(pl.col("period.number") == 3))
        .then(1200)
        .when((pl.col("lag_period") == pl.col("period.number") - 1).and_(pl.col("period.number") >= 4))
        .then(1200)
        .otherwise(pl.col("end.game_seconds_remaining"))
        .alias("end.game_seconds_remaining"),
    )

    return pbp_txt



from typing import Dict

import pandas as pd
import polars as pl

from sportsdataverse.dl_utils import download


def nhl_api_pbp(game_id: int, **kwargs) -> Dict:
    """nhl_api_pbp() - Pull the game by id. Data from API endpoints - `nhl/playbyplay`, `nhl/summary`

    Args:
        game_id (int): Unique game_id, can be obtained from nhl_schedule().

    Returns:
        Dict: Dictionary of game data with keys - "gameId", "plays", "boxscore", "header", "broadcasts",
         "videos", "playByPlaySource", "standings", "leaders", "seasonseries", "pickcenter", "againstTheSpread",
         "odds", "onIce", "gameInfo", "season"

    Example:
        `nhl_df = sportsdataverse.nhl.nhl_api_pbp(game_id=2021020079)`
    """
    # summary endpoint for pickcenter array
    summary_url = f"https://statsapi.web.nhl.com/api/v1/game/{game_id}/feed/live?site=en_nhl"
    summary_resp = download(summary_url, **kwargs)
    summary = summary_resp.json()
    pbp_txt = {"datetime": summary.get("gameData").get("datetime")}
    pbp_txt["game"] = summary.get("gameData").get("game")
    pbp_txt["players"] = summary.get("gameData").get("players")
    pbp_txt["status"] = summary.get("gameData").get("status")
    pbp_txt["teams"] = summary.get("gameData").get("teams")
    pbp_txt["venues"] = summary.get("gameData").get("venues")
    pbp_txt["gameId"] = summary.get("gameData").get("gamePk")
    pbp_txt["gameLink"] = summary.get("gameData").get("link")
    return pbp_txt


def nhl_api_schedule(start_date: str, end_date: str, return_as_pandas=False, **kwargs) -> pl.DataFrame:
    """nhl_api_schedule() - Pull the schedule by start and end date. Data from API endpoints - `nhl/schedule`

    Args:
        start_date (str): Start date to pull the NHL API schedule.
        end_date (str): End date to pull the NHL API schedule.
        return_as_pandas (bool): If True, returns a pandas dataframe. If False, returns a polars dataframe.

    Returns:
        pl.DataFrame: Polars dataframe containing the schedule for the requested seasons.

    Example:
        `nhl_sched_df = sportsdataverse.nhl.nhl_api_schedule(start_date=2021-10-23, end_date=2021-10-28)`
    """
    # summary endpoint for pickcenter array
    summary_url = "https://statsapi.web.nhl.com/api/v1/schedule"
    params = {"site": "en_nhl", "startDate": start_date, "endDate": end_date}
    summary_resp = download(summary_url, params=params, **kwargs)
    summary = summary_resp.json()
    pbp_txt = {"dates": summary.get("dates")}
    pbp_txt_games = pl.DataFrame()
    for date in pbp_txt["dates"]:
        game = pl.from_pandas(pd.json_normalize(date, record_path="games", meta=["date"]))
        pbp_txt_games = pl.concat([pbp_txt_games, game], how="vertical")
    return pbp_txt_games.to_pandas() if return_as_pandas else pbp_txt_games
